# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/facenet.ipynb.

# %% auto 0
__all__ = ['FaceNetInceptionResnetV1', 'facenet']

# %% ../nbs/facenet.ipynb 3
from facenet_pytorch import InceptionResnetV1, fixed_image_standardization

from .utils import *
from .siamese import *
from .pair_matching import *


class FaceNetInceptionResnetV1(InceptionResnetV1):
    """Like `pytorch_facenet`'s `InceptionResnetV1` but accepts standard float tensors"""
    def __init__(self, pretrained=True, classify=True):
        if pretrained is True:
            pretrained = 'vggface2'
        super().__init__(pretrained, classify)

    def forward(self, x):
        x = fixed_image_standardization(x * 255)
        return super().forward(x)


def facenet(pretrained=True) -> ThresholdSiamese:
    """A binary classifier matching pairs of facial images"""
    backbone = FaceNetInceptionResnetV1(pretrained, classify=False)
    backbone = cut_model_by_name(backbone, 'last_bn')  # skip normalization in case we want unnormalized features
    return ThresholdSiamese(backbone)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/pair_matching.ipynb.

# %% auto 0
__all__ = ['ThresholdSiamese']

# %% ../nbs/pair_matching.ipynb 3
import torch
from torch import nn
import numpy as np
from fastai.vision.all import *
from fastprogress.fastprogress import *

from .siamese import *
from .utils import *


class ThresholdSiamese(nn.Module):
    """Combines a deep similarity metric with a threshold to produce a binary classifier."""
    @delegates(DistanceSiamese)
    def __init__(self, *args, **kwargs):
        super().__init__()
        self.distance = DistanceSiamese(*args, **kwargs)
        self.threshold = Threshold()

    def forward(self, x):
        return self.threshold(self.distance(x))

    @delegates(DistanceSiamese.plot_distance_histogram)
    def plot_distance_histogram(self, *args, **kwargs):
        self.distance.plot_distance_histogram(*args, **kwargs)
        plt.axvline(self.threshold.t.item(), linestyle='--', color='grey', label='Threshold')

# %% ../nbs/pair_matching.ipynb 10
@patch
def fit_threshold(self: ThresholdSiamese, train_dl: DataLoader):
    """Picks a threshold that maximizes the accuracy on a dataloader"""
    self.eval().to(train_dl.device)
    with torch.no_grad():
        distances, targets = zip(*((self.distance(x), y) for x, y in progress_bar(train_dl, leave=False)))
        distances, targets = torch.cat(distances).as_subclass(Tensor), torch.cat(targets).as_subclass(Tensor)

        return self.threshold.fit(distances, targets) 



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/utils.ipynb.

# %% auto 0
__all__ = ['as_percentage', 'cut_model_by_name', 'MLP', 'Threshold', 'ExperimentalResults', 'RepeatedExperiment']

# %% ../nbs/utils.ipynb 3
from fastai.vision.all import *
from torchvision.models.feature_extraction import create_feature_extractor

from fastai_datasets.all import *

# %% ../nbs/utils.ipynb 4
def as_percentage(x, ndigits=2):
    return f'{round(x * 100, ndigits=ndigits)}%'

# %% ../nbs/utils.ipynb 5
def cut_model_by_name(model, cut):
    graph = create_feature_extractor(model, [cut])
    base_forward = graph.forward
    graph.forward = lambda *args, **kwargs: base_forward(*args, **kwargs)[cut]
    return graph

# %% ../nbs/utils.ipynb 7
class MLP(Module):
    """
    A Multilayer Perceptron comprized of linear layers with ReLU activations.
    """
    def __init__(self,
                 logits: Optional[int], # Number of logits. pass `None` to not include a logits layer after the hidden layers
                 hidden_depth=5, # Number of hidden layers
                 hidden_width=512, # Number of neurons in each hidden layer
                 features_dim=None  # Number of neurons in the last hidden layer. Pass `None` to use the same value as the other hidden layers
                 ):
        super().__init__()
        features_dim = features_dim or hidden_width

        def generate_hidden_layers():
            if hidden_depth == 0:
                return
            if hidden_depth >= 2:
                yield nn.LazyLinear(hidden_width)
                yield nn.ReLU()
                for _ in range(hidden_depth - 2):
                    yield nn.Linear(hidden_width, hidden_width)
                    yield nn.ReLU()
            yield nn.LazyLinear(features_dim)

        self.hidden_layers = nn.Sequential(*generate_hidden_layers())
        self.logits = nn.LazyLinear(logits) if logits else None

    def forward(self, x):
        x = x.flatten(start_dim=1)
        x = self.hidden_layers(x)
        if self.logits:
            x = self.logits(x)
        return x

# %% ../nbs/utils.ipynb 8
import torch
from torch import nn

from fastai.vision.all import *

class Threshold(nn.Module):
    """Classifies 1D inputs into 2 classes, based on whether they surpass a threshold or not"""
    def __init__(self) -> None:
        super().__init__()
        self.t = nn.Parameter(torch.zeros(1))

    def forward(self, x):
        x = x - self.t
        return torch.stack([x, -x], dim=-1)

# %% ../nbs/utils.ipynb 11
@patch
def fit(self: Threshold, x, y):
    """Picks a threshold that maximizes the empirical accuracy"""
    with torch.no_grad():
        def accuracy_for_threshold(t):
            self.t[0] = t
            return accuracy(self(x), y)

        threshold_candidates = np.arange(0.0, 4.0, 0.01)
        self.t[0], accuracy_score = max(((t, accuracy_for_threshold(t)) for t in threshold_candidates),
                                                key=lambda p: p[1])

        return self.t.item(), accuracy_score.item()


# %% ../nbs/utils.ipynb 15
from abc import ABC, abstractmethod
from dataclasses import dataclass

import matplotlib.pyplot as plt
    

@dataclass
class ExperimentalResults(object):
    "Provides various ways of examining the results of a `RepeatedExperiment`"
    stats: List[Any]

    @property
    def collated_stats(self):
        return {k: v.numpy() for k, v in default_collate(self.stats).items()}

    def plot_stats(self):
        fig, axs = plt.subplots(len(self.collated_stats), sharex=True)
        fig.tight_layout()
        for ax, (stat_name, stat_values) in zip(axs, self.collated_stats.items()):
            ax.set_title(stat_name)

            min_val, max_val = min(stat_values), max(stat_values)
            val_range = max_val - min_val
            ax.set_ylim(min_val - .1*val_range, max_val + .1*val_range)
            
            ax.bar(range(len(stat_values)), stat_values)

    @property
    def stat_means(self):
        return {k: v.mean() for k, v in self.collated_stats.items()}

    @property
    def stat_stds(self):
        return {k: v.std() for k, v in self.collated_stats.items()}


class RepeatedExperiment(ABC):
    "Runs multiple independent iterations of the same procedure, and combines the results"

    def __init__(self,
                 model: nn.Module,  # The model to be used in each iteration. Parameter are reset to their initial values before each iteration
                 data: List[Datasets],  # A list of `Datasets`, each representing a different iteration. A `Dataloaders` of the current `Datasets` is available via `self.dls`
                 seed: int = 0  # Used for reproducibility of results. Use `None` to avoid reproducibility
                 ):
        super().__init__()
        store_attr('model, data, seed')
    
    def run(self) -> ExperimentalResults:
        "Runs the experiment, returning the results as an `ExperimentalResults`"
        return ExperimentalResults(self._run())

    @return_list
    def _run(self):
        if self.seed is not None:
            set_seed(self.seed, reproducible=True)

        initial_state_dict = deepcopy(self.model.state_dict())

        splits = master_bar(self.data)
        for i, split in enumerate(splits):
            self.dls = split.dls()
            yield self.iteration()
            self.model.load_state_dict(initial_state_dict)

    @abstractmethod
    def iteration(self) -> Dict[str, Any]:
        pass



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/feature_space_plotting.ipynb.

# %% auto 0
__all__ = ['plot_dataset_embedding']

# %% ../nbs/feature_space_plotting.ipynb 3
import numpy as np
from fastai.vision.all import *
import matplotlib.pyplot as plt

# %% ../nbs/feature_space_plotting.ipynb 4
def plot_dataset_embedding(dataset: Datasets, feature_extractor, num_samples_per_class=300, normalize_features=False, *args, **kwargs):
    COLORS = ['#1f77b4', '#ff7f0e', '#2ca02c', '#d62728', '#9467bd', '#8c564b', '#e377c2', '#7f7f7f', '#bcbd22', '#17becf']

    feature_extractor.eval()

    with torch.no_grad():
        for c, ss in dataset.by_target.items():
            color = COLORS[dataset.vocab.o2i[c] % len(COLORS)]
            samples = ss.random_sub_dsets(num_samples_per_class).load()[0]
            embeddings = feature_extractor(samples)
            if normalize_features:
                embeddings = torch.nn.functional.normalize(embeddings, dim=1)

            _plot_cluster(embeddings.cpu().numpy(), color, label=c, *args, **kwargs)

    plt.legend()

def _plot_cluster(points: ndarray, color, *args, **kwargs):
    assert points.ndim == 2 and points.shape[1] == 2
    plt.scatter(*points.transpose(), color=color, *args, **kwargs)



# Autogenerated by nbdev

d = { 'settings': { 'branch': 'main',
                'doc_baseurl': '/similarity-learning',
                'doc_host': 'https://Irad-Zehavi.github.io',
                'git_url': 'https://github.com/Irad-Zehavi/similarity-learning',
                'lib_path': 'similarity_learning'},
  'syms': { 'similarity_learning.all': {},
            'similarity_learning.facenet': { 'similarity_learning.facenet.FaceNetInceptionResnetV1': ( 'facenet.html#facenetinceptionresnetv1',
                                                                                                       'similarity_learning/facenet.py'),
                                             'similarity_learning.facenet.FaceNetInceptionResnetV1.__init__': ( 'facenet.html#facenetinceptionresnetv1.__init__',
                                                                                                                'similarity_learning/facenet.py'),
                                             'similarity_learning.facenet.FaceNetInceptionResnetV1.forward': ( 'facenet.html#facenetinceptionresnetv1.forward',
                                                                                                               'similarity_learning/facenet.py'),
                                             'similarity_learning.facenet.facenet': ( 'facenet.html#facenet',
                                                                                      'similarity_learning/facenet.py')},
            'similarity_learning.feature_space_plotting': { 'similarity_learning.feature_space_plotting._plot_cluster': ( 'feature_space_plotting.html#_plot_cluster',
                                                                                                                          'similarity_learning/feature_space_plotting.py'),
                                                            'similarity_learning.feature_space_plotting.plot_dataset_embedding': ( 'feature_space_plotting.html#plot_dataset_embedding',
                                                                                                                                   'similarity_learning/feature_space_plotting.py')},
            'similarity_learning.pair_matching': { 'similarity_learning.pair_matching.ThresholdSiamese': ( 'pair_matching.html#thresholdsiamese',
                                                                                                           'similarity_learning/pair_matching.py'),
                                                   'similarity_learning.pair_matching.ThresholdSiamese.__init__': ( 'pair_matching.html#thresholdsiamese.__init__',
                                                                                                                    'similarity_learning/pair_matching.py'),
                                                   'similarity_learning.pair_matching.ThresholdSiamese.fit_threshold': ( 'pair_matching.html#thresholdsiamese.fit_threshold',
                                                                                                                         'similarity_learning/pair_matching.py'),
                                                   'similarity_learning.pair_matching.ThresholdSiamese.forward': ( 'pair_matching.html#thresholdsiamese.forward',
                                                                                                                   'similarity_learning/pair_matching.py'),
                                                   'similarity_learning.pair_matching.ThresholdSiamese.plot_distance_histogram': ( 'pair_matching.html#thresholdsiamese.plot_distance_histogram',
                                                                                                                                   'similarity_learning/pair_matching.py')},
            'similarity_learning.siamese': { 'similarity_learning.siamese.ContrastiveLoss': ( 'siamese.html#contrastiveloss',
                                                                                              'similarity_learning/siamese.py'),
                                             'similarity_learning.siamese.ContrastiveLoss.__call__': ( 'siamese.html#contrastiveloss.__call__',
                                                                                                       'similarity_learning/siamese.py'),
                                             'similarity_learning.siamese.ContrastiveLoss.__init__': ( 'siamese.html#contrastiveloss.__init__',
                                                                                                       'similarity_learning/siamese.py'),
                                             'similarity_learning.siamese.DistanceSiamese': ( 'siamese.html#distancesiamese',
                                                                                              'similarity_learning/siamese.py'),
                                             'similarity_learning.siamese.DistanceSiamese.__init__': ( 'siamese.html#distancesiamese.__init__',
                                                                                                       'similarity_learning/siamese.py'),
                                             'similarity_learning.siamese.DistanceSiamese._hist': ( 'siamese.html#distancesiamese._hist',
                                                                                                    'similarity_learning/siamese.py'),
                                             'similarity_learning.siamese.DistanceSiamese.forward': ( 'siamese.html#distancesiamese.forward',
                                                                                                      'similarity_learning/siamese.py'),
                                             'similarity_learning.siamese.DistanceSiamese.plot_distance_histogram': ( 'siamese.html#distancesiamese.plot_distance_histogram',
                                                                                                                      'similarity_learning/siamese.py'),
                                             'similarity_learning.siamese.normalized_squared_euclidean_distance': ( 'siamese.html#normalized_squared_euclidean_distance',
                                                                                                                    'similarity_learning/siamese.py')},
            'similarity_learning.utils': { 'similarity_learning.utils.ExperimentalResults': ( 'utils.html#experimentalresults',
                                                                                              'similarity_learning/utils.py'),
                                           'similarity_learning.utils.ExperimentalResults.collated_stats': ( 'utils.html#experimentalresults.collated_stats',
                                                                                                             'similarity_learning/utils.py'),
                                           'similarity_learning.utils.ExperimentalResults.plot_stats': ( 'utils.html#experimentalresults.plot_stats',
                                                                                                         'similarity_learning/utils.py'),
                                           'similarity_learning.utils.ExperimentalResults.stat_means': ( 'utils.html#experimentalresults.stat_means',
                                                                                                         'similarity_learning/utils.py'),
                                           'similarity_learning.utils.ExperimentalResults.stat_stds': ( 'utils.html#experimentalresults.stat_stds',
                                                                                                        'similarity_learning/utils.py'),
                                           'similarity_learning.utils.MLP': ('utils.html#mlp', 'similarity_learning/utils.py'),
                                           'similarity_learning.utils.MLP.__init__': ( 'utils.html#mlp.__init__',
                                                                                       'similarity_learning/utils.py'),
                                           'similarity_learning.utils.MLP.forward': ( 'utils.html#mlp.forward',
                                                                                      'similarity_learning/utils.py'),
                                           'similarity_learning.utils.RepeatedExperiment': ( 'utils.html#repeatedexperiment',
                                                                                             'similarity_learning/utils.py'),
                                           'similarity_learning.utils.RepeatedExperiment.__init__': ( 'utils.html#repeatedexperiment.__init__',
                                                                                                      'similarity_learning/utils.py'),
                                           'similarity_learning.utils.RepeatedExperiment._run': ( 'utils.html#repeatedexperiment._run',
                                                                                                  'similarity_learning/utils.py'),
                                           'similarity_learning.utils.RepeatedExperiment.iteration': ( 'utils.html#repeatedexperiment.iteration',
                                                                                                       'similarity_learning/utils.py'),
                                           'similarity_learning.utils.RepeatedExperiment.run': ( 'utils.html#repeatedexperiment.run',
                                                                                                 'similarity_learning/utils.py'),
                                           'similarity_learning.utils.Threshold': ('utils.html#threshold', 'similarity_learning/utils.py'),
                                           'similarity_learning.utils.Threshold.__init__': ( 'utils.html#threshold.__init__',
                                                                                             'similarity_learning/utils.py'),
                                           'similarity_learning.utils.Threshold.fit': ( 'utils.html#threshold.fit',
                                                                                        'similarity_learning/utils.py'),
                                           'similarity_learning.utils.Threshold.forward': ( 'utils.html#threshold.forward',
                                                                                            'similarity_learning/utils.py'),
                                           'similarity_learning.utils.as_percentage': ( 'utils.html#as_percentage',
                                                                                        'similarity_learning/utils.py'),
                                           'similarity_learning.utils.cut_model_by_name': ( 'utils.html#cut_model_by_name',
                                                                                            'similarity_learning/utils.py')}}}



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/siamese.ipynb.

# %% auto 0
__all__ = ['normalized_squared_euclidean_distance', 'ContrastiveLoss', 'DistanceSiamese']

# %% ../nbs/siamese.ipynb 3
from torch import nn
from torch.nn.functional import normalize

from fastai.vision.all import *


def normalized_squared_euclidean_distance(x1, x2):
    r"""
    Squared Euclidean distance over normalized vectors:
    $$\left\| \frac{x_1}{\|x_1\|}-\frac{x_2}{\|x_2\|} \right\|^2 $$
    """
    assert x1.dim() <= 2
    assert x2.dim() <= 2
    x1 = normalize(x1, dim=-1)
    x2 = normalize(x2, dim=-1)
    return (x1 - x2).pow(2).sum(dim=-1)


class ContrastiveLoss(BaseLoss):
    @delegates(nn.HingeEmbeddingLoss)
    def __init__(self, **kwargs):
        super().__init__(loss_cls=nn.HingeEmbeddingLoss, **kwargs)

    def __call__(self, input, target):
        return super().__call__(input, 2*target-1)  # hinge_embedding_loss expects targets to be 1 or -1


# @dataclass
class DistanceSiamese(Module):
    """Outputs the distance between two inputs in feature space"""
    def __init__(self, 
                 backbone: Module,  # embeds inputs in a feature space
                 distance_metric = normalized_squared_euclidean_distance):
        self.backbone = backbone
        self.distance_metric = distance_metric

    def forward(self, x):
        f1, f2 = L(x).map(self.backbone).map(nn.Flatten())
        return self.distance_metric(f1, f2)


# %% ../nbs/siamese.ipynb 4
from matplotlib.ticker import PercentFormatter
from fastprogress.fastprogress import *
import numpy as np

@patch
def plot_distance_histogram(self: DistanceSiamese, pairs_dls: Union[TfmdDL, Dict[str, TfmdDL]], label='Distance'):
    """Plots a histogram of intra-class and inter-class distances"""    
    if isinstance(pairs_dls, DataLoader):
        pairs_dls = {
            'Intra-Class': pairs_dls.by_target['Same'],
            'Inter-Class': pairs_dls.by_target['Not Same']
        }

    self.eval()

    with torch.no_grad():
        for label, dl in pairs_dls.items():
            self._hist(dl, label)
    
    plt.legend()
    plt.xlabel(label)


@patch
def _hist(self: DistanceSiamese, dl, label):
    self.to(dl.device)
    processed_batches = [self(x) for x, _ in progress_bar(dl, leave=False)]
    distances = torch.cat(processed_batches).cpu().numpy()
    weights = np.ones_like(distances) / len(distances)  # normalization for percentage ticks
    plt.hist(distances, label=label, alpha=.5, edgecolor='black', lw=1, weights=weights)
    plt.gca().yaxis.set_major_formatter(PercentFormatter(1))
